Index: callerid.c
===================================================================
RCS file: /usr/cvsroot/asterisk/callerid.c,v
retrieving revision 1.19
diff -u -r1.19 callerid.c
--- callerid.c	23 Jul 2004 00:27:26 -0000	1.19
+++ callerid.c	19 Aug 2004 17:31:16 -0000
@@ -20,6 +20,7 @@
 #include <stdlib.h>
 #include <unistd.h>
 #include <math.h>
+#include <ctype.h>
 #include <asterisk/ulaw.h>
 #include <asterisk/alaw.h>
 #include <asterisk/frame.h>
@@ -145,6 +146,64 @@
 		*number = NULL;
 	else
 		*number = cid->number;
+}
+
+void callerid_get_dtmf(char *cidstring, char *number, int *flags)
+{
+	int i;
+	int code;
+
+	if (strlen(cidstring) < 2) {
+		ast_log(LOG_DEBUG, "No cid detected\n");
+		*flags = CID_UNKNOWN_NUMBER;
+		return;
+	}
+	
+	/* Detect protocol and special types */
+	if (cidstring[0] == 'B') {
+		/* Handle special codes */
+		code = atoi(&cidstring[1]);
+		if (code == 0)
+			*flags = CID_UNKNOWN_NUMBER;
+		else if (code == 10) 
+			*flags = CID_PRIVATE_NUMBER;
+		else
+			ast_log(LOG_DEBUG, "Unknown DTMF code %d\n", code);
+	} else if (cidstring[0] == 'D' && cidstring[2] == '#') {
+		/* .DK special code */
+		if (cidstring[1] == '1')
+			*flags = CID_PRIVATE_NUMBER;
+		if (cidstring[1] == '2' || cidstring[1] == '3')
+			*flags = CID_UNKNOWN_NUMBER;
+	} else if (cidstring[0] == 'D' || cidstring[0] == 'A') {
+		/* "Standard" callerid */
+		for (i = 1; i < strlen(cidstring); i++ ) {
+			if (cidstring[i] == 'C' || cidstring[i] == '#')
+				break;
+			if (isdigit(cidstring[i]))
+				number[i-1] = cidstring[i];
+			else
+				ast_log(LOG_DEBUG, "Unknown CID digit '%c'\n",
+					cidstring[i]);
+		}
+		number[i-1] = 0;
+	} else if (isdigit(cidstring[0])) {
+		/* It begins with a digit, so we parse it as a number and hope
+		 * for the best */
+		ast_log(LOG_WARNING, "Couldn't detect start-character. CID "
+			"parsing might be unreliable\n");
+		for (i = 0; i < strlen(cidstring); i++) {
+			if (isdigit(cidstring[i]))
+                                number[i] = cidstring[i];
+			else
+				break;
+		}
+		number[i] = 0;
+	} else {
+		ast_log(LOG_DEBUG, "Unknown CID protocol, start digit '%c'\n", 
+			cidstring[0]);
+		*flags = CID_UNKNOWN_NUMBER;
+	}
 }
 
 int ast_gen_cas(unsigned char *outbuf, int sendsas, int len, int codec)
Index: channels/chan_zap.c
===================================================================
RCS file: /usr/cvsroot/asterisk/channels/chan_zap.c,v
retrieving revision 1.322
diff -u -r1.322 chan_zap.c
--- channels/chan_zap.c	12 Aug 2004 22:09:18 -0000	1.322
+++ channels/chan_zap.c	19 Aug 2004 17:31:17 -0000
@@ -862,7 +862,8 @@
 		"Ringer Off",
 		"Hook Transition Complete",
 		"Bits Changed",
-		"Pulse Start"
+	        "Pulse Start",
+	        "Polarity Reversal"
 };
 
 static struct {
@@ -3329,6 +3330,8 @@
 			case SIG_SF_FEATD:
 			case SIG_SF_FEATDMF:
 			case SIG_SF_FEATB:
+				if (ast->_state == AST_STATE_PRERING)
+					ast_setstate(ast, AST_STATE_RING);
 				if ((ast->_state == AST_STATE_DOWN) || (ast->_state == AST_STATE_RING)) {
 					if (option_debug)
 						ast_log(LOG_DEBUG, "Ring detected\n");
@@ -3584,7 +3587,7 @@
 			default:
 				break;
 			}
-			break;
+			break;		
 		default:
 			ast_log(LOG_DEBUG, "Dunno what to do with event %d on channel %d\n", res, p->channel);
 	}
@@ -4480,6 +4483,7 @@
 	char exten2[AST_MAX_EXTENSION]="";
 	unsigned char buf[256];
 	char cid[256];
+	char dtmfcid[300];
 	char dtmfbuf[300];
 	struct callerid_state *cs;
 	char *name=NULL, *number=NULL;
@@ -4986,6 +4990,54 @@
 	case SIG_FXSLS:
 	case SIG_FXSGS:
 	case SIG_FXSKS:
+		/* If we are in a pre-ring state, handle whatever needs 
+		 * handling, and wait for ring */
+		if (chan->_state == AST_STATE_PRERING) {
+			if (p->use_callerid == CIDTYPE_DTMF) {
+				int i = 0;
+				ast_log(LOG_DEBUG, "Receiving DTMF cid on "
+					"channel %s\n", chan->name);
+				zt_setlinear(p->subs[index].zfd, 0);
+				res = 2000;
+				for (;;) {
+					struct ast_frame *f;
+					res = ast_waitfor(chan, res);
+					if (res <= 0) {
+						ast_log(LOG_WARNING, "DTMFCID timed out waiting for ring. "
+							"Exiting simple switch\n");
+						ast_hangup(chan);
+						return NULL;
+					} 
+					f = ast_read(chan);
+					if (f->frametype == AST_FRAME_DTMF) {
+						dtmfbuf[i++] = f->subclass;
+						ast_log(LOG_DEBUG, "CID got digit '%c'\n", f->subclass);
+						res = 2000;
+					}
+					ast_frfree(f);
+					if (chan->_state == AST_STATE_RING ||
+					    chan->_state == AST_STATE_RINGING) 
+						break; /* Got ring */
+				}
+				dtmfbuf[i] = 0;
+				zt_setlinear(p->subs[index].zfd, p->subs[index].linear);
+				/* Got cid and ring. */
+				ast_log(LOG_DEBUG, "CID got string '%s'\n", dtmfbuf);
+				callerid_get_dtmf(dtmfbuf, dtmfcid, &flags);
+				ast_log(LOG_DEBUG, "CID is '%s', flags %d\n", 
+					dtmfcid, flags);
+				number = dtmfcid;
+			} else {
+				ast_log(LOG_WARNING, "Channel %s in prering "
+					"state, but I have nothing to do. "
+					"Terminating simple switch, should be "
+					"restarted by the actual ring.\n", 
+					chan->name);
+				ast_hangup(chan);
+				return NULL;
+			}
+		}
+
 #ifdef ZAPATA_PRI
 		if (p->pri) {
 			/* This is a GR-303 trunk actually.  Wait for the first ring... */
@@ -5017,7 +5069,7 @@
 			}
 		}
 #endif
-		if (p->use_callerid) {
+		if (p->use_callerid == CIDTYPE_FSK) {
 			cs = callerid_new();
 			if (cs) {
 #if 1
@@ -5376,6 +5428,26 @@
 			return -1;
 		}
 		break;
+	case ZT_EVENT_POLARITY:
+		switch(i->sig) {
+		case SIG_FXSLS:
+		case SIG_FXSKS:
+		case SIG_FXSGS:
+			if (i->use_callerid == CIDTYPE_DTMF) {
+				ast_verbose(VERBOSE_PREFIX_2 "Starting DTMF "
+					    "CID detection on channel %d\n",
+					    i->channel);
+				chan = zt_new(i, AST_STATE_PRERING, 0, SUB_REAL, 0, 0);
+				if (chan && ast_pthread_create(&threadid, &attr, ss_thread, chan)) {
+					ast_log(LOG_WARNING, "Unable to start simple switch thread on channel %d\n", i->channel);
+				}
+			}
+			break;
+		default:
+			ast_log(LOG_WARNING, "handle_init_event detected "
+				"polarity reversal on non-FXO (SIG_FXS) "
+				"interface %d\n", i->channel);
+		}
 	}
 	return 0;
 }
@@ -8778,7 +8850,12 @@
 			ringc = v->value;
 			sscanf(ringc, "%d,%d,%d", &drings.ringnum[2].ring[0], &drings.ringnum[2].ring[1], &drings.ringnum[2].ring[2]);
 		} else if (!strcasecmp(v->name, "usecallerid")) {
-			use_callerid = ast_true(v->value);
+			if (!strcasecmp(v->value, "dtmf")) 
+				use_callerid = CIDTYPE_DTMF;
+			else if (!strcasecmp(v->value, "fsk"))
+				use_callerid = CIDTYPE_FSK;
+			else if (ast_true(v->value))
+				use_callerid = CIDTYPE_FSK;
 		} else if (!strcasecmp(v->name, "threewaycalling")) {
 			threewaycalling = ast_true(v->value);
 		} else if (!strcasecmp(v->name, "cancallforward")) {
@@ -9436,7 +9513,12 @@
 			ringc = v->value;
 			sscanf(ringc, "%d,%d,%d", &drings.ringnum[2].ring[0], &drings.ringnum[2].ring[1], &drings.ringnum[2].ring[2]);
 		} else if (!strcasecmp(v->name, "usecallerid")) {
-			use_callerid = ast_true(v->value);
+			if (!strcasecmp(v->value, "dtmf")) 
+				use_callerid = CIDTYPE_DTMF;
+			else if (!strcasecmp(v->value, "fsk")) 
+				use_callerid = CIDTYPE_FSK;
+			else if (ast_true(v->value)) 
+				use_callerid = CIDTYPE_FSK;
 		} else if (!strcasecmp(v->name, "threewaycalling")) {
 			threewaycalling = ast_true(v->value);
 		} else if (!strcasecmp(v->name, "transfer")) {
Index: include/asterisk/callerid.h
===================================================================
RCS file: /usr/cvsroot/asterisk/include/asterisk/callerid.h,v
retrieving revision 1.6
diff -u -r1.6 callerid.h
--- include/asterisk/callerid.h	7 Aug 2004 15:51:11 -0000	1.6
+++ include/asterisk/callerid.h	19 Aug 2004 17:31:17 -0000
@@ -24,6 +24,10 @@
 #define CID_UNKNOWN_NAME		(1 << 2)
 #define CID_UNKNOWN_NUMBER		(1 << 3)
 
+#define CIDTYPE_FSK  1
+#define CIDTYPE_DTMF 2
+
+
 #define AST_LIN2X(a) ((codec == AST_FORMAT_ALAW) ? (AST_LIN2A(a)) : (AST_LIN2MU(a)))
 #define AST_XLAW(a) ((codec == AST_FORMAT_ALAW) ? (AST_ALAW(a)) : (AST_MULAW(a)))
 
@@ -83,6 +87,15 @@
  * Returns nothing.
  */
 void callerid_get(struct callerid_state *cid, char **number, char **name, int *flags);
+
+//! Get and parse DTMF-based callerid 
+/*!
+ * \param cidstring The actual transmitted string.
+ * \param number The cid number is returned here.
+ * \param flags The cid flags are returned here.
+ * This function parses DTMF callerid.
+ */
+void callerid_get_dtmf(char *cidstring, char *number, int *flags);
 
 //! Free a callerID state
 /*!
Index: include/asterisk/channel.h
===================================================================
RCS file: /usr/cvsroot/asterisk/include/asterisk/channel.h,v
retrieving revision 1.54
diff -u -r1.54 channel.h
--- include/asterisk/channel.h	6 Aug 2004 14:43:25 -0000	1.54
+++ include/asterisk/channel.h	19 Aug 2004 17:31:17 -0000
@@ -338,6 +338,8 @@
 #define AST_STATE_BUSY  	7		
 /*! Digits (or equivalent) have been dialed while offhook */
 #define AST_STATE_DIALING_OFFHOOK	8
+/*! Channel has detected an incoming call and is waiting for ring */
+#define AST_STATE_PRERING       9
 
 /* Bits 16-32 of state are reserved for flags */
 /*! Do not transmit voice data */
